<?php

	/*
	$url = "http://example.com/";
	$drumkitImageDir = "drumkitImages/"; // this is a subdirectory under the dir hydrogen.php resides

	$drumkit_list[ "K-27_Trash_Kit" ][ "author" ] = "wolke";
	$drumkit_list[ "K-27_Trash_Kit" ][ "info" ] = "A real trash kit!";
	*/

?>
